/*
    $Id: aes.c 292 2010-02-14 07:37:34Z irfangul82 $
    Simple file and directory encryption
    Copyright (C) 2010  Irfan Gulamali

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <stdlib.h>
#include <string.h>
#include "glob.h"
#include "aes.h"

/*
 * http://en.wikipedia.org/wiki/Rijndael_Galois_field#Rijndael.27s_finite_field
 Multiply two numbers in the GF(2^8) finite field defined
 by the polynomial x^8 + x^4 + x^3 + x + 1
 }*/
static unsigned char gmul(unsigned char a, unsigned char b)
{
    unsigned char p = 0;
    int counter;
    int hi_bit_set;
    for (counter = 0; counter < 8; counter++)
    {
        if ((b & 1) == 1)
            p ^= a;
        hi_bit_set = (a & 0x80);
        a <<= 1;
        if (hi_bit_set == 0x80)
            a ^= 0x1b; /* x^8 + x^4 + x^3 + x + 1 */
        b >>= 1;
    }
    return p;
}

void initAESCTX(aesctx **ctx, char K[32])
{
    (*ctx) = malloc(sizeof(aesctx));
    memset((*ctx)->K_SCH, 0, sizeof((*ctx)->K_SCH));
    memcpy(&((*ctx)->K_SCH), K, 32);
    expandKey(*ctx);
}

void destroyAESCTX(aesctx **ctx)
{
    memset((*ctx), 0, sizeof(aesctx));
    free(*ctx);
}

void encryptAES256(aesctx *ctx)
{
    int ctr = 0;
    addRoundKey(ctx, &(ctx->K_SCH[(ctr) * 4]));
    ctr = 1;

    while (ctr < (NR))
    {
        subBytes(&(ctx->BLK[0]));
        shiftRows(&(ctx->BLK[0]));
        mixColumns(&(ctx->BLK[0]));
        addRoundKey(ctx, (int *) &(ctx->K_SCH[(ctr) * 4]));
        ctr++;
    }
    subBytes(&(ctx->BLK[0]));
    shiftRows(&(ctx->BLK[0]));
    addRoundKey(ctx, (int *) &(ctx->K_SCH[(ctr) * 4]));
}

void decryptAES256(aesctx *ctx)
{
    int ctr = 0;
    addRoundKey(ctx, &(ctx->K_SCH[59 - ((ctr + 1) * 4) + 1]));
    ctr = 1;

    while (ctr < (NR))
    {
        InvshiftRows(&(ctx->BLK[0]));
        InvsubBytes(&(ctx->BLK[0]));
        addRoundKey(ctx, (int *) &(ctx->K_SCH[59 - ((ctr + 1) * 4) + 1]));
        InvmixColumns(&(ctx->BLK[0]));
        ctr++;
    }
    InvshiftRows(&(ctx->BLK[0]));
    InvsubBytes(&(ctx->BLK[0]));
    addRoundKey(ctx, (int *) &(ctx->K_SCH[59 - ((ctr + 1) * 4) + 1]));
}

/* AES */
static void expandKey(aesctx *ctx)
{
    unsigned int temp;
    unsigned int temp2;
    int ctr = 0;

    //fix endianness
    if (L_ENDIAN())
    {
        while (ctr < 8)
        {
            REVBYTE32(ctx->K_SCH[ctr]);
            ctr++;
        }
    }

    while (ctr < (NB * (NR + 1)))
    {
        temp = ctx->K_SCH[ctr - 1];
        if ((ctr % NK) == 0)
        {
            temp2 = Rcon[ctr / NK];
            if (L_ENDIAN())
                REVBYTE32(temp2);
            temp = subWord((rotWord(temp))) ^ temp2;
        }
        else if (((ctr - 4) % NK) == 0)
        {
            temp = subWord(temp);
        }
        ctx->K_SCH[ctr] = (ctx->K_SCH[ctr - NK]) ^ temp;
        ctr++;
    }
}

static unsigned int rotWord(unsigned int word)
{
    unsigned int rotword;
    memcpy(&rotword, &word, 4);
    rotword = ((rotword << 8) ^ (rotword >> 24));
    return rotword;
}

static unsigned int subWord(unsigned int word)
{
    unsigned int subword;
    unsigned char *wptr = (char *) &word;
    unsigned char sword[4];
    char ctr = 0;
    while (ctr < 4)
    {
        memcpy(&sword[ctr], &S_BOX[(int) *wptr], 1);
        wptr++;
        ctr++;
    }
    memcpy(&subword, &sword, 4);
    return subword;
}

static void subBytes(unsigned char *state)
{
    int ctr = 0;
    int tmp = 0;
    while (ctr < 16)
    {
        tmp = *state;
        memcpy(state, &S_BOX[tmp], 1);
        state++;
        ctr++;
    }
}

static void shiftRows(unsigned char *state)
{
    /*
     * Array arrives as single dimension 16 byte array
     * assumption was that 4x4 block was composed as such:
     * 63cab7040953d051cd60e0e7ba70e18c
     * and processed for first row as such:
     * state[0] - state[3] (row 1)
     * state[4] - state[7] (row 2)
     * state[8] - state[11] (row 3)
     * state[12] - state[15] (row 4)
     *
     * but is actually processed as such (possible endian issue):
     *
     * 63 ca b7 04
     * 09 53 d0 51
     * cd 60 e0 e7
     * ba 70 e1 8c
     *
     * 63 53 e0 8c
     * 09 60 e1 04
     * cd 70 b7 51
     * ba ca d0 e7
     */

    char *st = 0;
    char tmp;
    st = state;

    tmp = st[1];
    st[1] = st[5];
    st[5] = st[9];
    st[9] = st[13];
    st[13] = tmp;

    tmp = st[2];
    st[2] = st[10];
    st[10] = tmp;
    tmp = st[6];
    st[6] = st[14];
    st[14] = tmp;

    tmp = st[15];
    st[15] = st[11];
    st[11] = st[7];
    st[7] = st[3];
    st[3] = tmp;

}

static void mixColumns(unsigned char *state)
{
    int ctr = 0;
    unsigned char *st_h;
    unsigned char st[4];
    st_h = state;

    while (ctr < 15)
    {
        st[0] = gmul(st_h[ctr], 0x02) ^ gmul(st_h[ctr + 1], 0x03) ^ st_h[ctr
                + 2] ^ st_h[ctr + 3];
        st[1] = st_h[ctr] ^ gmul(st_h[ctr + 1], 0x02) ^ gmul(st_h[ctr + 2],
                0x03) ^ st_h[ctr + 3];
        st[2] = st_h[ctr] ^ st_h[ctr + 1] ^ gmul(st_h[ctr + 2], 0x02) ^ gmul(
                    st_h[ctr + 3], 0x03);
        st[3] = gmul(st_h[ctr], 0x03) ^ st_h[ctr + 1] ^ st_h[ctr + 2] ^ gmul(
                    st_h[ctr + 3], 0x02);

        st_h[ctr] = st[0];
        st_h[ctr + 1] = st[1];
        st_h[ctr + 2] = st[2];
        st_h[ctr + 3] = st[3];

        ctr += 4;
    }
}

static void addRoundKey(aesctx *ctx, unsigned int *k_sch)
{
    unsigned char *d = 0;
    unsigned char *k = 0;
    int ctr = 0;

    d = &(ctx->BLK[0]);

    //could be a potential problem on big endian
    k = (char *) k_sch;

    while (ctr < 16)
    {
        d[3] ^= (*k);
        k++;
        d[2] ^= (*k);
        k++;
        d[1] ^= (*k);
        k++;
        d[0] ^= (*k);
        k++;
        d++;
        d++;
        d++;
        d++;
        ctr += 4;

    }
    //-------------------------------

}

static void InvsubBytes(unsigned char *state)
{
    int ctr = 0;
    int tmp = 0;
    while (ctr < 16)
    {
        tmp = *state;
        memcpy(state, &S_BOXINV[tmp], 1);
        state++;
        ctr++;
    }
}

static void InvshiftRows(unsigned char *state)
{
    /* shiftrows below:
     *
     * 63 ca b7 04
     * 09 53 d0 51
     * cd 60 e0 e7
     * ba 70 e1 8c
     *
     * 63 53 e0 8c
     * 09 60 e1 04
     * cd 70 b7 51
     * ba ca d0 e7
     */

    char *st = 0;
    char tmp;
    st = state;

    tmp = st[13];
    st[13] = st[9];
    st[9] = st[5];
    st[5] = st[1];
    st[1] = tmp;

    tmp = st[2];
    st[2] = st[10];
    st[10] = tmp;
    tmp = st[6];
    st[6] = st[14];
    st[14] = tmp;

    tmp = st[3];
    st[3] = st[7];
    st[7] = st[11];
    st[11] = st[15];
    st[15] = tmp;
}

static void InvmixColumns(unsigned char *state)
{
    int ctr = 0;
    unsigned char *st_h;
    unsigned char st[4];
    st_h = state;

    while (ctr < 15)
    {
        st[0] = gmul(st_h[ctr], 0x0e) ^ gmul(st_h[ctr + 1], 0x0b) ^ gmul(
                    st_h[ctr + 2], 0x0d) ^ gmul(st_h[ctr + 3], 0x09);
        st[1] = gmul(st_h[ctr], 0x09) ^ gmul(st_h[ctr + 1], 0x0e) ^ gmul(
                    st_h[ctr + 2], 0x0b) ^ gmul(st_h[ctr + 3], 0x0d);
        st[2] = gmul(st_h[ctr], 0x0d) ^ gmul(st_h[ctr + 1], 0x09) ^ gmul(
                    st_h[ctr + 2], 0x0e) ^ gmul(st_h[ctr + 3], 0x0b);
        st[3] = gmul(st_h[ctr], 0x0b) ^ gmul(st_h[ctr + 1], 0x0d) ^ gmul(
                    st_h[ctr + 2], 0x09) ^ gmul(st_h[ctr + 3], 0x0e);

        st_h[ctr] = st[0];
        st_h[ctr + 1] = st[1];
        st_h[ctr + 2] = st[2];
        st_h[ctr + 3] = st[3];

        ctr += 4;
    }
}

